@@ -1,10 +1,10 @@
 /****************************************************************************
-Copyright ReportLab Europe Ltd. 2000-2019 see license.txt for license details
+Copyright ReportLab Europe Ltd. 2000-2021 see license.txt for license details
  ****************************************************************************/
 #define PY_SSIZE_T_CLEAN
 #include <Python.h>
-#if PY_MAJOR_VERSION >= 3
-#	define isPy3
+#if PY_MAJOR_VERSION <= 2
+#	error Only Python-3.x is supported
 #endif
 #include <stdio.h>
 #include <stdlib.h>
@@ -22,46 +22,53 @@ Copyright ReportLab Europe Ltd. 2000-2019 see license.txt for license details
 #include "stdio16.h"
 #include "version.h"
 #include "namespaces.h"
-#define VERSION "2.2.0"
+#define VERSION "3.0.1"
 #define MAX_DEPTH 256
 #if PY_VERSION_HEX < 0x02050000
 #	define Py_ssize_t int
 #endif
 
+struct __FILE16 {	/*we only need this part*/
+	void *handle;
+	int handle2, handle3;
+	};
+#define __F16A(_f,_a) (((struct __FILE16*)_f)->_a)
+#define __F16H(_f) __F16A(_f,handle)
+#define __F16H2(_f) __F16A(_f,handle2)
+#define __F16H3(_f) __F16A(_f,handle3)
+
 static	int	g_byteorder;
 static	char *g_encname;
 
-struct module_state {
-	PyObject *moduleError;
-	PyObject *moduleVersion;
-	PyObject *RXPVersion;
-	PyObject *commentTagName;
-	PyObject *piTagName;
-	PyObject *CDATATagName;
-	PyObject *recordLocation;
-	PyObject *parser_flags;
-	PyObject *parser;
-	};
-
 typedef struct {
 	PyObject_HEAD
 	PyObject		*warnCB, *eoCB, *ugeCB, *srcName, *fourth, *__instance_module__;
 	int				flags[2];
 	} pyRXPParser;
 
-#ifdef isPy3
-#	define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
-#	define MSTATE(m,n) GETSTATE(m)->n
-#	define PPSTATE(p,n) MSTATE(((pyRXPParser *)p)->__instance_module__,n)
-#	define PDSTATE(pd,n) PPSTATE(((ParserDetails*)pd)->__self__,n)
-#	define PSTATE(p,n) PDSTATE(((Parser)p)->warning_callback_arg,n)
-#	define PyInt_FromLong	PyLong_FromLong
-#	define PyInt_AsLong	PyLong_AsLong
-#	define staticforward static
-#	define statichere static
-	/*cmp(a,b) = (a > b) - (a < b) in Python3*/
-#	define PyObject_Cmp(a,b,i) *i=(PyObject_RichCompareBool(a, b, Py_GT)-PyObject_RichCompareBool(a, b, Py_LT))
-#	define PyNumber_Int	PyNumber_Long
+#define MSTATE(m,n) _moduleGetAttr(m,stringize(n))
+#define PPSTATE(p,n) MSTATE(((pyRXPParser *)p)->__instance_module__,n)
+#define PDSTATE(pd,n) PPSTATE(((ParserDetails*)pd)->__self__,n)
+#define PSTATE(p,n) PDSTATE(((Parser)p)->warning_callback_arg,n)
+
+static PyObject *moduleError = NULL;
+
+static PyObject *_moduleGetAttr(PyObject *m, const char *a) {
+    PyObject *v = NULL;
+
+    v = PyDict_GetItemString(PyModule_GetDict(m), a);
+    if (! v) {
+        PyErr_Format(PyExc_AttributeError, "Module '%s' has no attribute '%s'.", PyModule_GetName(m), a);
+        return NULL;
+    	}
+	return v;
+}
+
+#define PyInt_FromLong	PyLong_FromLong
+#define PyInt_AsLong	PyLong_AsLong
+/*cmp(a,b) = (a > b) - (a < b) in Python3*/
+#define PyObject_Cmp(a,b,i) *i=(PyObject_RichCompareBool(a, b, Py_GT)-PyObject_RichCompareBool(a, b, Py_LT))
+#define PyNumber_Int	PyNumber_Long
 PyObject *RLPy_FindMethod(PyMethodDef *ml, PyObject *self, const char* name){
 	for(;ml->ml_name!=NULL;ml++)
 		if(name[0]==ml->ml_name[0] && strcmp(name+1,ml->ml_name+1)==0) return PyCFunction_New(ml, self);
@@ -72,67 +79,36 @@ PyObject *RLPy_FindMethod(PyMethodDef *ml, PyObject *self, const char* name){
 		}
 	return NULL;
 	}
-#	define Py_FindMethod RLPy_FindMethod
-#	define KEY2STR(key) PyUnicode_AsUTF8(key)
+#define Py_FindMethod RLPy_FindMethod
+#define KEY2STR(key) PyUnicode_AsUTF8(key)
+#if	defined(DEBUG_PYRXP) && defined(WIN32)
+	include <crtdbg.h>
+#endif
+#if 0 && defined(DEBUG_PYRXP)
+#	define RLPy_VISIT(o,n) fprintf(stderr,"VISIT o->" stringize(n) "=%p |o->" stringize(n) "|=%d\n", o->n, o->n ? Py_REFCNT(o->n) : -1317);if(o->n && Py_REFCNT(o->n)>0) Py_VISIT(o->n)
+#	define RLPy_CLEAR(o,n) fprintf(stderr,"CLEAR o->" stringize(n) "=%p |o->" stringize(n) "|=%d\n", o->n, o->n ? Py_REFCNT(o->n) : -1317); Py_CLEAR(o->n)
 #else
-	static	PyObject *g_module;
-	static struct module_state _state;
-#	define GETSTATE(m) (&_state)
-#	define MSTATE(m,n) GETSTATE(m)->n
-#	define PPSTATE(p,n) MSTATE(p,n)
-#	define PDSTATE(pd,n) MSTATE(pd,n)
-#	define PSTATE(p,n) MSTATE(p,n)
-#	include "bytesobject.h"
-#	ifndef PyVarObject_HEAD_INIT
-#		define PyVarObject_HEAD_INIT(type, size) \
-			PyObject_HEAD_INIT(type) size,
-#	endif
-#	ifndef Py_TYPE
-#		define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
-#	endif
-#	define PyBytes_AS_STRING	PyString_AS_STRING
-#	define PyBytes_AsString	PyString_AsString
-#	define PyBytes_GET_SIZE		PyString_GET_SIZE
-#	define KEY2STR(key) PyBytes_AsString(key)
+#	define RLPy_VISIT(o,n) if(o->n && Py_REFCNT(o->n)>0) Py_VISIT(o->n)
+#	define RLPy_CLEAR(o,n) Py_CLEAR(o->n)
 #endif
 
-#if CHAR_SIZE==16
-#	define MODULENAME "pyRXPU"
-#	define UTF8DECL ,int utf8
-#	define UTF8PASS ,utf8
-#	define PYNSNAME(nsed, name) PyNSName(nsed,name,utf8)
+#define MODULENAME "pyRXPU"
+#define UTF8DECL ,int utf8
+#define UTF8PASS ,utf8
+#define PYNSNAME(nsed, name) PyNSName(nsed,name,utf8)
 PyObject* _PYSTRING(const Char* s, int utf8)
 {
 	int	lens = (int)Strlen(s);
-#if defined(Py_UNICODE_WIDE)
 	PyObject *u, *x;
 	u = PyUnicode_DecodeUTF16((char*)s,2*lens,NULL,NULL);
 	if(!utf8 || !u) return u;
 	x = PyUnicode_AsUTF8String(u);
 	Py_DECREF(u);
 	return x;
-#else
-	return utf8 ? PyUnicode_EncodeUTF8((Py_UNICODE*)s, lens, NULL)
-				: PyUnicode_FromUnicode((Py_UNICODE*)s, lens);
-#endif
 }
 #	define PYSTRING(s) _PYSTRING(s,utf8)
-PyObject* PYSTRING8(const char* s)
-{
-	return PyUnicode_DecodeUTF8((const char*)s, (int)strlen(s), NULL);
-}
 #	define EmptyCharStr (Char*)"\0"
 #	define FMTCHAR "u"
-#else
-#	define MODULENAME "pyRXP"
-#	define UTF8DECL
-#	define UTF8PASS
-#	define PYNSNAME(nsed, name) PyNSName(nsed,name)
-#	define PYSTRING(s) PyString_FromString(s)
-#	define PYSTRING8(s) PyString_FromString(s)
-#	define EmptyCharStr (Char*)""
-#	define FMTCHAR "s"
-#endif
 PyObject* PyNSName(NSElementDefinition nsed, const Char *name UTF8DECL){
 	Char		*t, *ns;
 	Namespace	NS;
@@ -408,9 +384,6 @@ static struct {char* k;long v;} flag_vals[]={
 #endif
 #define __GetFlag(p, flag) \
   ((((flag) < 32) ? ((p)->flags[0] & (1u << (flag))) : ((p)->flags[1] & (1u << ((flag)-32))))!=0)
-#ifdef	_DEBUG
-#	define Py_REFCOUNT(op) ((op)->ob_refcnt)
-#endif
 
 typedef	struct {
 		Parser		p;
@@ -427,9 +400,7 @@ typedef	struct {
 #if	CHAR_SIZE==16
 		int			utf8;
 #endif
-#ifdef	isPy3
 		pyRXPParser*	__self__;	/*the associated parser object*/
-#endif
 		} ParserDetails;
 
 #define PDGetItem pd->GetItem
@@ -466,7 +437,7 @@ static	PyObject* _getSrcInfo(ParserDetails *pd)
 	PyObject *t = PyTuple_New(3);
 	const char *name = EntityDescription(s->entity);
 	int lnum, cnum;
-	PyTuple_SET_ITEM(t,0,PYSTRING8(name));
+	PyTuple_SET_ITEM(t,0,PyUnicode_FromString(name));
 	switch(SourceLineAndChar(s, &lnum, &cnum)){
 		case 0:
 		case 1:
@@ -658,9 +629,7 @@ static InputSource entity_open(Entity e, void *info)
 	PyObject	*eoCB = pd->eoCB, *text=NULL, *tmp;
 
 	if(e->type==ET_external){
-		PyObject	*arglist,*result;
-		arglist = Py_BuildValue("(s)",e->systemid);	/*NB 8 bit*/
-		result = PyEval_CallObject(eoCB, arglist);
+		PyObject	*result = PyObject_CallFunction(eoCB,"s",e->systemid);
 		if(result){
 			int isTuple=PyTuple_Check(result), isBytes=PyBytes_Check(result);
 			if(!(isBytes||isTuple)&&PyUnicode_Check(result)){
@@ -678,13 +647,13 @@ static InputSource entity_open(Entity e, void *info)
 					if(PyUnicode_Check(tmp)){
 						tmp = PyUnicode_AsEncodedString(tmp,"utf8","strict");
 						if(!tmp){
-							PyErr_SetString(PDSTATE(pd,moduleError),"eoCB could not convert tuple URI (element 0) from unicode");
+							PyErr_SetString(moduleError,"eoCB could not convert tuple URI (element 0) from unicode");
 L_err0:						Py_DECREF(result);
 							return NULL;
 							}
 						}
 					else if(!PyBytes_Check(tmp)){
-						PyErr_SetString(PDSTATE(pd,moduleError),"eoCB could not convert tuple URI (element 0) from unknown type");
+						PyErr_SetString(moduleError,"eoCB could not convert tuple URI (element 0) from unknown type");
 						goto L_err0;
 						}
 					e->systemid = strdup8(PyBytes_AS_STRING(tmp));
@@ -701,7 +670,6 @@ L_err0:						Py_DECREF(result);
 		else {
 			PyErr_Clear();
 			}
-		Py_DECREF(arglist);
 		}
 	if(text){
 		int textlen;
@@ -714,13 +682,13 @@ L_err0:						Py_DECREF(result);
 				text = tmp;
 				}
 			else {
-				PyErr_SetString(PDSTATE(pd,moduleError),"eoCB could not convert tuple text value");
+				PyErr_SetString(moduleError,"eoCB could not convert tuple text value");
 				Py_DECREF(text);
 				return NULL;
 				}
 			}
 		else if(!PyBytes_Check(text)){
-			PyErr_SetString(PDSTATE(pd,moduleError),"eoCB returned tuple with non-text value");
+			PyErr_SetString(moduleError,"eoCB returned tuple with non-text value");
 			Py_DECREF(text);
 			return NULL;
 			}
@@ -746,14 +714,14 @@ void PyErr_FromStderr(Parser p, char *msg){
 	if(p->errbuf) Fprintf(Stderr,"%s\n", p->errbuf);
 	Fprintf(Stderr,"%s\n", msg);
 	buf[((struct _FILE16*)Stderr)->handle2] = 0;
-	PyErr_SetString(PSTATE(p,moduleError),buf);
+	PyErr_SetString(moduleError,buf);
 #else
 	PyObject* t;
 	if(p->errbuf) Fprintf(Stderr,"%s\n", p->errbuf);
 	Fprintf(Stderr,"%s\n", msg);
 	t = PyUnicode_DecodeUTF16((const char *)buf, (Py_ssize_t)(((struct _FILE16*)Stderr)->handle2), NULL, &g_byteorder);
 	if(!t) return;
-	PyErr_SetObject(PSTATE(p,moduleError),t);
+	PyErr_SetObject(moduleError,t);
 	Py_DECREF(t);
 #endif
 }
@@ -829,29 +797,30 @@ PyObject *ProcessSource(Parser p, InputSource source)
 static void myWarnCB(XBit bit, void *info)
 {
 	ParserDetails*	pd=(ParserDetails*)info;
-	PyObject	*arglist;
-	PyObject	*result;
-	FILE16		*str;
-	char		buf[512];
+	PyObject	*a, *result;
+	FILE16		*sf;
 
 	pd->warnErr++;
 	if(pd->warnCB==Py_None) return;
 
-	str = MakeFILE16FromString(buf,sizeof(buf)-1,"w");
-	_ParserPerror(str, pd->p, bit);
-	Fclose(str);
+	sf = MakeStringFILE16("w");
+	_ParserPerror(sf, pd->p, bit);
 #if	CHAR_SIZE==16
+	a = PyUnicode_DecodeUTF16((char *)__F16H(sf), __F16H2(sf), NULL, NULL);
+#else
+	a = PyBytes_FromStringAndSize((char *)__F16H(sf),__F16H2(sf));
+#endif
+#if 0
 	{
-	struct _FILE16 {
-		void *handle;
-		int handle2, handle3;
-		};
-	buf[((struct _FILE16*)str)->handle2] = 0;
+	int i;
+	fprintf(stderr,"myWarnCB: a=%p sf=%p sfh=%p sfh2=%d sfh3=%d\n b=b'", a, sf,__F16H(sf),__F16H2(sf),__F16H3(sf));
+	for(i=0;i<__F16H2(sf);i++) fprintf(stderr,"\\x%02x",((unsigned char *)__F16H(sf))[i]);
+	fprintf(stderr,"'\n");
 	}
 #endif
-	arglist = Py_BuildValue("(" FMTCHAR ")",buf);
-	result = PyEval_CallObject(pd->warnCB, arglist);
-	Py_DECREF(arglist);
+	Fclose(sf);
+	result = PyObject_CallFunctionObjArgs(pd->warnCB, a, NULL);
+	Py_XDECREF(a);
 	if(result){
 		Py_DECREF(result);
 		}
@@ -864,7 +833,6 @@ static void myWarnCB(XBit bit, void *info)
 static Char *myUGECB(Char *name, int namelen, void *info)
 {
 	ParserDetails*	pd=(ParserDetails*)info;
-	PyObject	*arglist;
 	PyObject	*result;
 	PyObject	*uname;
 	PyObject	*bytes;
@@ -876,10 +844,8 @@ static Char *myUGECB(Char *name, int namelen, void *info)
 	if(pd->ugeCB==Py_None) return r;
 	uname = PyUnicode_DecodeUTF16((const char *)name, (Py_ssize_t)(sizeof(Char)*namelen), NULL, &g_byteorder);
 	if(!uname) return r;
-	arglist = Py_BuildValue("(O)",uname);
+	result = PyObject_CallFunctionObjArgs(pd->ugeCB, uname, NULL);
 	Py_DECREF(uname);
-	result = PyEval_CallObject(pd->ugeCB, arglist);
-	Py_DECREF(arglist);
 	if(result){
 		if(PyBytes_Check(result)){
 			/*if we see bytes we asssume it's utf8 encoded*/
@@ -899,7 +865,7 @@ static Char *myUGECB(Char *name, int namelen, void *info)
 						r = (Char*)Malloc((int)(sz+sizeof(Char)));
 						if(r){
 							memcpy((char *)r,s,sz);
-							for(ir=0;ir<sizeof(Char);ir++) ((char *)r)[sz+ir]=0;
+							for(ir=0;ir<(int)sizeof(Char);ir++) ((char *)r)[sz+ir]=0;
 							}
 						}
 					Py_DECREF(bytes);
@@ -1009,12 +975,11 @@ static PyObject* pyRXPParser_parse(pyRXPParser* xself, PyObject* args, PyObject*
 	ParserDetails	CB;
 	Parser		p;
 	Entity		e;
+	Dtd			pdtd;
 	pyRXPParser	dummy = *xself;
 	pyRXPParser*	self = &dummy;
 	memset(&CB,0,sizeof(CB));
-#ifdef	isPy3
 	CB.__self__ = self;
-#endif
 	if(self->warnCB) Py_INCREF(self->warnCB);
 	if(self->eoCB) Py_INCREF(self->eoCB);
 	if(self->ugeCB) Py_INCREF(self->ugeCB);
@@ -1099,9 +1064,10 @@ static PyObject* pyRXPParser_parse(pyRXPParser* xself, PyObject* args, PyObject*
 	source = SourceFromFILE16(PyBytes_AsString(self->srcName),f);
 	retVal = ProcessSource(p,source);
 	e = source->entity; /*used during FreeParser closing source!*/
+	pdtd = p->dtd;
 	Fclose(Stderr);
-	FreeDtd(p->dtd);
 	FreeParser(p);
+	FreeDtd(pdtd);
 	FreeEntity(e);
 	deinit_parser();
 L_1:
@@ -1156,19 +1122,19 @@ static PyObject* pyRXPParser_getattr(pyRXPParser *self, char *name)
 	}
 
 static int pyRXPParser_traverse(pyRXPParser *self, visitproc visit, void *arg){
-	Py_VISIT(self->srcName);
-	Py_VISIT(self->warnCB);
-	Py_VISIT(self->eoCB);
-	Py_VISIT(self->fourth);
-	Py_VISIT(self->__instance_module__);
+	RLPy_VISIT(self,srcName);
+	RLPy_VISIT(self,warnCB);
+	RLPy_VISIT(self,eoCB);
+	RLPy_VISIT(self,fourth);
+	RLPy_VISIT(self,__instance_module__);
 	return 0;
 	}
 static int pyRXPParser_clear(pyRXPParser* self){
-	Py_CLEAR(self->srcName);
-	Py_CLEAR(self->warnCB);
-	Py_CLEAR(self->eoCB);
-	Py_CLEAR(self->fourth);
-	Py_CLEAR(self->__instance_module__);
+	RLPy_CLEAR(self,srcName);
+	RLPy_CLEAR(self,warnCB);
+	RLPy_CLEAR(self,eoCB);
+	RLPy_CLEAR(self,fourth);
+	RLPy_CLEAR(self,__instance_module__);
 	return 0;
 	}
 
@@ -1178,12 +1144,11 @@ static int pyRXPParser_dealloc(pyRXPParser* self){
 	return 0;
 	}
 
-#ifdef isPy3
 static struct PyModuleDef moduleDef;
-#endif
 static int pyRXPParser_init(pyRXPParser* self, PyObject* args, PyObject* kw)
 {
 	Py_ssize_t	i;
+	PyObject *ob;
 
 	if(!PyArg_ParseTuple(args, ":Parser")) return -1;
 	Py_XDECREF(self->warnCB);
@@ -1192,16 +1157,24 @@ static int pyRXPParser_init(pyRXPParser* self, PyObject* args, PyObject* kw)
 	Py_XDECREF(self->fourth);
 	Py_XDECREF(self->srcName);
 	Py_XDECREF(self->__instance_module__);
-	self->warnCB = self->eoCB = self->ugeCB = self->fourth = self->srcName = NULL;
-#ifdef isPy3
-	self->__instance_module__ = PyState_FindModule(&moduleDef);
-#else
-	self->__instance_module__ = g_module;
-#endif
+	self->warnCB = self->eoCB = self->ugeCB = self->fourth = self->srcName = self->__instance_module__ = NULL;
+
+	/*this hackery is to find the module that we're part of*/
+	ob = PyImport_GetModuleDict();	/*this is per interpeter*/
+	if (!ob) {
+		PyErr_SetString(moduleError,"Cannot find the systemn modules dict!");
+		return -1;
+	}
+	self->__instance_module__ = PyDict_GetItemString(ob,"pyRXPU");
+	if(!self->__instance_module__) {
+		PyErr_SetString(moduleError,"Cannot find the pyRXPU module object!");
+		return -1;
+	}
+
 	Py_INCREF(self->__instance_module__);
 	if(!(self->srcName=PyBytes_FromString("[unknown]"))){
-		PyErr_SetString(MSTATE(self->__instance_module__,moduleError),"Internal error, memory limit reached!");
-Lfree:	pyRXPParser_dealloc(self);
+		PyErr_SetString(moduleError,"Internal error, memory limit reached!");
+Lfree:	/*pyRXPParser_dealloc(self);*/
 		return -1;
 		}
 	for(i=0;flag_vals[i].k;i++)
@@ -1256,140 +1229,76 @@ static PyTypeObject pyRXPParserType = {
 	0,										/*tp_dictoffset*/
 	(initproc)pyRXPParser_init,				/*tp_init*/
 	0,										/*tp_alloc*/
-	0,										/*tp_new*/
+	PyType_GenericNew,						/*tp_new*/
 	};
 
+static int module_exec(PyObject *m)
+{
+	PyObject *parser_flags, *t;
+	int i;
 
-#if	defined(_DEBUG) && defined(WIN32)
-#	include <crtdbg.h>
-#endif
-#ifdef isPy3
-static int _traverse(PyObject *m, visitproc visit, void *arg) {
-	struct module_state *st = GETSTATE(m);
-	Py_VISIT(st->moduleError);
-	Py_VISIT(st->moduleVersion);
-	Py_VISIT(st->RXPVersion);
-	Py_VISIT(st->piTagName);
-	Py_VISIT(st->recordLocation);
-	Py_VISIT(st->parser_flags);
-	Py_VISIT(st->parser);
-	return 0;
-	}
+	/* Add some attributes to the module */
+	if(PyModule_AddStringConstant(m, "version", VERSION)<0) goto fail;
+	if(PyModule_AddStringConstant(m, "RXPVersion", rxp_version_string)<0) goto fail;
+	if(PyModule_AddStringConstant(m, "recordLocation", "recordLocation")<0) goto fail;
+	if(PyModule_AddStringConstant(m, "piTagName", "<?")<0) goto fail;
+	if(PyModule_AddStringConstant(m, "commentTagName", "<!--")<0) goto fail;
+	if(PyModule_AddStringConstant(m, "CDATATagName", "<![CDATA[")<0) goto fail;
+    if (moduleError == NULL) {
+		moduleError = PyErr_NewException(MODULENAME ".error",NULL,NULL);
+		if(!moduleError) goto fail;
+		}
+    Py_INCREF(moduleError);
+    if(PyModule_AddObject(m, "error", moduleError)<0) goto fail;
+
+
+	parser_flags = PyDict_New();
+	if(!parser_flags)goto fail;
+	for(i=0;flag_vals[i].k;i++){
+		t=PyInt_FromLong(flag_vals[i].v);
+		if(!t)goto fail;
+		PyDict_SetItemString(parser_flags, flag_vals[i].k, t);
+		Py_DECREF(t);
+		}
+    if(PyModule_AddObject(m, "parser_flags", parser_flags)<0) goto fail;
+
+    pyRXPParserType.tp_base = &PyBaseObject_Type;
+	if(PyType_Ready(&pyRXPParserType)<0) goto fail;
+	if(PyModule_AddObject(m,"Parser", (PyObject *)&pyRXPParserType)<0) goto fail;
 
-static int _clear(PyObject *m) {
-	struct module_state *st = GETSTATE(m);
-	Py_CLEAR(st->moduleError);
-	Py_CLEAR(st->moduleVersion);
-	Py_CLEAR(st->RXPVersion);
-	Py_CLEAR(st->piTagName);
-	Py_CLEAR(st->recordLocation);
-	Py_CLEAR(st->parser_flags);
-	Py_CLEAR(st->parser);
 	return 0;
-	}
+fail:
+    Py_XDECREF(m);
+    return -1;
+}
+
+static struct PyModuleDef_Slot module_slots[] = {
+    {Py_mod_exec, module_exec},
+    {0, NULL},
+};
+
 static struct PyModuleDef moduleDef = {
 	PyModuleDef_HEAD_INIT,
 	MODULENAME,
 	__DOC__,
-	sizeof(struct module_state),
+	0,
+	NULL,
+	module_slots,
 	NULL,
 	NULL,
-	_traverse,
-	_clear,
 	NULL
 	};
-#	if CHAR_SIZE==16
-#		define MODULEINIT PyInit_pyRXPU
-#	else
-#		define MODULEINIT PyInit_pyRXP
-#	endif
-#	define OK_RET m
-#	define ERR_RET NULL
-#	define CREATE_MODULE() PyModule_Create(&moduleDef);PyState_AddModule(m,&moduleDef)
-PyMODINIT_FUNC MODULEINIT(void)
-#else
-#	if CHAR_SIZE==16
-#		define MODULEINIT initpyRXPU
-#	else
-#		define MODULEINIT initpyRXP
-#	endif
-#	define OK_RET
-#	define ERR_RET
-#	define CREATE_MODULE() Py_InitModule3(MODULENAME, NULL, __DOC__)
-DL_EXPORT(void) MODULEINIT(void)
-#endif
+
+PyMODINIT_FUNC PyInit_pyRXPU(void)
 {
-	PyObject *m=NULL, *t, *moduleVersion=NULL, *RXPVersion=NULL, *moduleError=NULL,
-			 *piTagName=NULL, *commentTagName=NULL, *CDATATagName=NULL, *recordLocation=NULL,
-			 *parser_flags=NULL;
-	int	i;
 	g_byteorder = InternalCharacterEncoding==CE_UTF_16B?1:-1;
 	g_encname = g_byteorder==1?"utf_16_be":"utf_16_le";
-
-#if	defined(_DEBUG) && defined(WIN32)
+#if	defined(DEBUG_PYRXP) && defined(WIN32)
+	int	i;
 	i = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
 	i |= _CRTDBG_CHECK_ALWAYS_DF;
 	_CrtSetDbgFlag(i);
 #endif
 
-	pyRXPParserType.tp_new = PyType_GenericNew;
-	if(PyType_Ready(&pyRXPParserType)<0)goto err;	/*set up the types by hand*/
-
-	/* Create the module and add the functions */
-	m = CREATE_MODULE();
-	if(!m)goto err;
-
-#ifndef isPy3
-	g_module = m;
-#endif
-
-	/* Add some symbolic constants to the module */
-	moduleVersion = PyBytes_FromString(VERSION);
-	if(!moduleVersion)goto err;
-	RXPVersion = PyBytes_FromString(rxp_version_string);
-	if(!RXPVersion)goto err;
-	moduleError = PyErr_NewException(MODULENAME ".error",NULL,NULL);
-	if(!moduleError)goto err;
-	piTagName = PYSTRING8("<?");
-	if(!piTagName)goto err;
-	commentTagName = PYSTRING8("<!--");
-	if(!commentTagName)goto err;
-	CDATATagName = PYSTRING8("<![CDATA[");
-	if(!CDATATagName)goto err;
-	recordLocation = PyBytes_FromString("recordLocation");
-	if(!recordLocation)goto err;
-	parser_flags = PyDict_New();
-	if(!parser_flags)goto err;
-	for(i=0;flag_vals[i].k;i++){
-		t=PyInt_FromLong(flag_vals[i].v);
-		if(!t)goto err;
-		PyDict_SetItemString(parser_flags, flag_vals[i].k, t);
-		Py_DECREF(t);
-		}
-
-	/*add in the docstring*/
-#define ADD2MODULE(n,o) PyModule_AddObject(m,n,o);MSTATE(m,o)=o
-	ADD2MODULE("version", moduleVersion);
-	ADD2MODULE("RXPVersion", RXPVersion);
-	ADD2MODULE("error",moduleError);
-	ADD2MODULE("piTagName", piTagName);
-	ADD2MODULE("commentTagName", commentTagName);
-	ADD2MODULE("CDATATagName", CDATATagName);
-	ADD2MODULE("recordLocation", recordLocation);
-	ADD2MODULE("parser_flags", parser_flags);
-	Py_INCREF((PyObject *)&pyRXPParserType);
-	PyModule_AddObject(m,"Parser", (PyObject *)&pyRXPParserType);
-	MSTATE(m,parser) = (PyObject *)&pyRXPParserType;
-	return OK_RET;
-err:
-	Py_XDECREF(moduleVersion);
-	Py_XDECREF(RXPVersion);
-	Py_XDECREF(moduleError);
-	Py_XDECREF(piTagName);
-	Py_XDECREF(commentTagName);
-	Py_XDECREF(CDATATagName);
-	Py_XDECREF(recordLocation);
-	Py_XDECREF(parser_flags);
-	Py_XDECREF(m);
-	return ERR_RET;
+    return PyModuleDef_Init(&moduleDef);
 }
